/*************************************************************************\
*                  Copyright (C) Jan Murawski, 2012.                      *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Public License as published                  *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
\*************************************************************************/

#include "NatTranslator.h"
#include "Logger.h"
#include "helper.h"

#include "unistd.h"
#include <linux/ip.h>
#include <linux/udp.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <string.h>

NatTranslator* NatTranslator::translator = 0;

NatTranslator::NatTranslator() 
{
    pthread_mutex_init(&cs_mutex, NULL);
    this->start_unattanded();
}

NatTranslator::~NatTranslator() 
{
}

NatTranslator* NatTranslator::getTranslator()
{
    if( translator == 0 )
        translator = new NatTranslator;
    return translator;
}

void NatTranslator::execute()
{
    //
    // clean up old connections each 5 seconds
    //
    while(true)
    {
        pthread_mutex_lock(&cs_mutex);
        
        time_t now = time(0);
        list<rule*>::iterator itr;
        for(itr = rules.begin(); itr != rules.end();)
        {
            if ( (*itr)->lastUsed + 60*3 < now )
            {
                if( (*itr)->sockfd != 0 )
                    close( (*itr)->sockfd );
                delete *itr;
                itr=rules.erase(itr);
            }
            else
              ++itr;
        }
        
        pthread_mutex_unlock(&cs_mutex);
        sleep(5);
    }
}

void NatTranslator::translateOutgoing(char* buffer)
{
    pthread_mutex_lock(&cs_mutex);
    
    struct iphdr* ip = (struct iphdr*)buffer;
    if( ip->protocol == IPPROTO_ICMP )
    {
        bool match = false;
        list<rule*>::iterator itr;
        for(itr = rules.begin(); itr != rules.end();)
        {
            if( (*itr)->protocol == IPPROTO_ICMP && 
                (*itr)->orig_src.ip == ip->saddr && 
                (*itr)->dst.ip == ip->daddr )
            {
                match = true;
                (*itr)->lastUsed = time(0);
                break;
            }
            ++itr;
        }
        
        if( !match )
        {
            rule* r = new rule;
            r->lastUsed = time(0);
            r->orig_src.ip = ip->saddr;
            r->dst.ip = ip->daddr;
            r->protocol = IPPROTO_ICMP;
            r->sockfd = 0;
            rules.push_back(r);
        }
        
        ip->saddr = 0;
    }
    else if( ip->protocol == IPPROTO_TCP )
    {
        struct tcphdr* tcp = (struct tcphdr*) (buffer + sizeof(struct iphdr));
        
        rule* selectedRule;
        bool match = false;
        list<rule*>::iterator itr;
        for(itr = rules.begin(); itr != rules.end();)
        {
            if( (*itr)->protocol == IPPROTO_TCP && 
                (*itr)->orig_src.ip == ip->saddr && 
                (*itr)->orig_src.port == tcp->source &&
                (*itr)->dst.ip == ip->daddr && 
                (*itr)->dst.port == tcp->dest )
            {
                match = true;
                (*itr)->lastUsed = time(0);
                selectedRule = *itr;
                break;
            }
            ++itr;
        }
        
        if( !match )
        {
            rule* r = new rule;
            r->lastUsed = time(0);
            r->orig_src.ip = ip->saddr;
            r->orig_src.port = tcp->source;
            r->dst.ip = ip->daddr;
            r->dst.port = tcp->dest;
            r->protocol = IPPROTO_TCP;
            
            /*
             * hack: create socket to save port
             */
            
            // create socket
            int sockfd = socket(AF_INET, SOCK_STREAM, 0);
            if (sockfd < 0) 
                Logger::error("ERROR opening socket");
            
            // bind
            struct sockaddr_in serv_addr;
            bzero((char *) &serv_addr, sizeof(serv_addr));
            serv_addr.sin_family = AF_INET;
            serv_addr.sin_addr.s_addr = htonl (INADDR_ANY); 
            serv_addr.sin_port = 0;
            if (bind(sockfd, (struct sockaddr *) &serv_addr,
                     sizeof(serv_addr)) < 0) 
                     Logger::error("ERROR on binding");
            
            // get bound local port
            sockaddr_in newAddr;
            int sin_size = sizeof(sockaddr_in);
            int nRetVal = getsockname(sockfd, (sockaddr *)&newAddr, (socklen_t *)&sin_size);
            
            r->sockfd = sockfd;
            r->src.port = newAddr.sin_port;
            
            /*
             * hack: create socket, to get local outgoing interface
             */
            
            // create socket
            sockfd = socket(AF_INET, SOCK_DGRAM, 0);
            
            // connect it
            struct sockaddr_in addr;
            memset(&addr, 0, sizeof(struct sockaddr_in));
            addr.sin_family      = AF_INET;
            addr.sin_addr.s_addr = ntohl(ip->daddr);
            addr.sin_port        = 0;
            connect(sockfd, (struct sockaddr *) &addr, sizeof(addr));

            // retrieve local iface
            sin_size = sizeof(sockaddr_in);
            nRetVal = getsockname(sockfd, (sockaddr *)&newAddr, (socklen_t *)&sin_size);
            close(sockfd);
            
            r->src.ip = newAddr.sin_addr.s_addr;
            
            rules.push_back(r);
            selectedRule = r;
        }
        
        ip->saddr = 0;
        ip->saddr = selectedRule->src.ip;
        tcp->source = selectedRule->src.port;
        tcp->check = 0;
        tcp->check = get_tcp_checksum(ip, tcp);
    }
    else // UDP
    {
        struct udphdr* udp = (struct udphdr*) (buffer + sizeof(struct iphdr));
        
        rule* selectedRule;
        bool match = false;
        list<rule*>::iterator itr;
        for(itr = rules.begin(); itr != rules.end();)
        {
            if( (*itr)->protocol == IPPROTO_UDP && 
                (*itr)->orig_src.ip == ip->saddr && 
                (*itr)->orig_src.port == udp->source &&
                (*itr)->dst.ip == ip->daddr && 
                (*itr)->dst.port == udp->dest )
            {
                match = true;
                (*itr)->lastUsed = time(0);
                selectedRule = *itr;
                break;
            }
            ++itr;
        }
        
        if( !match )
        {
            rule* r = new rule;
            r->lastUsed = time(0);
            r->orig_src.ip = ip->saddr;
            r->orig_src.port = udp->source;
            r->dst.ip = ip->daddr;
            r->dst.port = udp->dest;
            r->protocol = IPPROTO_UDP;
            
            /*
             * hack: create socket to save port
             */
            
            // create socket
            int sockfd = socket(AF_INET,SOCK_DGRAM, IPPROTO_UDP);
            if (sockfd < 0) 
                Logger::error("ERROR opening socket");
            
            // bind
            struct sockaddr_in serv_addr;
            bzero((char *) &serv_addr, sizeof(serv_addr));
            serv_addr.sin_family = AF_INET;
            serv_addr.sin_addr.s_addr = htonl (INADDR_ANY); 
            serv_addr.sin_port = 0;
            if (bind(sockfd, (struct sockaddr *) &serv_addr,
                     sizeof(serv_addr)) < 0) 
                     Logger::error("ERROR on binding");
            
            // get bound local port
            sockaddr_in newAddr;
            int sin_size = sizeof(sockaddr_in);
            int nRetVal = getsockname(sockfd, (sockaddr *)&newAddr, (socklen_t *)&sin_size);
            
            r->sockfd = sockfd;
            r->src.port = newAddr.sin_port;
            
            /*
             * hack: create socket, to get local outgoing interface
             */
            
            // create socket
            sockfd = socket(AF_INET, SOCK_DGRAM, 0);
            
            // connect it
            struct sockaddr_in addr;
            memset(&addr, 0, sizeof(struct sockaddr_in));
            addr.sin_family      = AF_INET;
            addr.sin_addr.s_addr = ntohl(ip->daddr);
            addr.sin_port        = 0;
            connect(sockfd, (struct sockaddr *) &addr, sizeof(addr));

            // retrieve local iface
            sin_size = sizeof(sockaddr_in);
            nRetVal = getsockname(sockfd, (sockaddr *)&newAddr, (socklen_t *)&sin_size);
            close(sockfd);
            
            r->src.ip = newAddr.sin_addr.s_addr;
            
            rules.push_back(r);
            selectedRule = r;
        }
        
        ip->saddr = 0;
        ip->saddr = selectedRule->src.ip;
        udp->source = selectedRule->src.port;
        udp->check = 0;
        udp->check = get_udp_checksum(ip, udp);
    }
    
    pthread_mutex_unlock(&cs_mutex);
}

int NatTranslator::translateIncomingICMP(char* buffer, int nr)
{
    int total = 0;
    struct iphdr* ip = (struct iphdr*)buffer;
    pthread_mutex_lock(&cs_mutex);
    
    //int debug = rules.size();
    list<rule*>::iterator itr;
    
    for(itr = rules.begin(); itr != rules.end();)
    {
        if( (*itr)->protocol == IPPROTO_ICMP && 
            (*itr)->dst.ip == ip->saddr )
        {
            ++total;
            if( total == nr +1 )
            {
                (*itr)->lastUsed = time(0);    
                ip->daddr = (*itr)->orig_src.ip;
            }
        }
        ++itr;
    }
    
    pthread_mutex_unlock(&cs_mutex);
    
    return total;
}

bool NatTranslator::translateIncomingTCP(char* buffer)
{
    struct iphdr* ip = (struct iphdr*)buffer;
    struct tcphdr* tcp = (struct tcphdr*) (buffer + sizeof(struct iphdr));
    pthread_mutex_lock(&cs_mutex);
    
    list<rule*>::iterator itr;
    bool match = false;
    for(itr = rules.begin(); itr != rules.end();)
    {
        if( (*itr)->protocol == IPPROTO_TCP &&
            (*itr)->dst.ip == ip->saddr &&
            (*itr)->dst.port == tcp->source &&
            (*itr)->src.port == tcp->dest  )
        {
            (*itr)->lastUsed = time(0);

            ip->daddr = (*itr)->orig_src.ip;
            tcp->dest = (*itr)->orig_src.port;
            tcp->check = 0;
            tcp->check = get_tcp_checksum(ip, tcp);
            
            match = true;
            break;
        }
        ++itr;
    }
    
    pthread_mutex_unlock(&cs_mutex);
    return match;
}

bool NatTranslator::translateIncomingUDP(char* buffer)
{
    struct iphdr* ip = (struct iphdr*)buffer;
    struct udphdr* udp = (struct udphdr*) (buffer + sizeof(struct iphdr));
    pthread_mutex_lock(&cs_mutex);
    
    list<rule*>::iterator itr;
    bool match = false;
    for(itr = rules.begin(); itr != rules.end();)
    {
        if( (*itr)->protocol == IPPROTO_UDP &&
            (*itr)->dst.ip == ip->saddr &&
            (*itr)->dst.port == udp->source &&
            (*itr)->src.port == udp->dest  )
        {
            (*itr)->lastUsed = time(0);

            ip->daddr = (*itr)->orig_src.ip;
            udp->dest = (*itr)->orig_src.port;
            udp->check = 0;
            udp->check = get_udp_checksum(ip, udp);
            
            match = true;
            break;
        }
        ++itr;
    }
    
    pthread_mutex_unlock(&cs_mutex);
    return match;
}
