/*************************************************************************\
*                  Copyright (C) Jan Murawski, 2012.                      *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Public License as published                  *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
\*************************************************************************/

#include "tunnelServer.h"
#include "helper.h"
#include "NatTranslator.h"

#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <limits.h>
#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/if_ether.h>

TunnelServer::TunnelServer()
{
    this->start();
}

void TunnelServer::execute()
{
     int sockfd, newsockfd;
     struct sockaddr_in serv_addr, cli_addr;

     // create socket
     sockfd = socket(AF_INET, SOCK_STREAM, 0);
     if (sockfd < 0) 
         Logger::error("ERROR opening socket");
     
     // set SO_REUSEADDR on a socket to true (1):
    int optval = 1;
    setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof optval);

     
     // bind
     bzero((char *) &serv_addr, sizeof(serv_addr));
     serv_addr.sin_family = AF_INET;
     inet_aton("127.0.0.1", &(serv_addr.sin_addr));
     serv_addr.sin_port = htons(1337);
     if (bind(sockfd, (struct sockaddr *) &serv_addr,
              sizeof(serv_addr)) < 0) 
              Logger::error("ERROR on binding");
     
     // and listen
     listen(sockfd,5);
     
     socklen_t clilen = sizeof(cli_addr);
     // accept connections
     for(;;)
     {
         newsockfd = accept(sockfd, 
                     (struct sockaddr *) &cli_addr, 
                     &clilen);
         if (newsockfd < 0)
         {
                Logger::error("ERROR on accept");
         }

         Logger::log("Accepted tunnel\n");
         new Connection(newsockfd);
     }
}


TunnelServer::Connection::Connection(int sockfd) 
{
    this->tunnelSock = sockfd;
    this->start_unattanded();
    
}

TunnelServer::Connection::~Connection()
{
    close(tunnelSock);
    close(rawSockIn);
    close(rawSockOut);
}


void TunnelServer::Connection::execute()
{
    //
    // raw socket (caputer ehternet frames, strip ethernet header, only ip packets)
    //
    if ((rawSockIn = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_IP))) == -1)
    //if ((rawSockIn = socket(PF_PACKET, SOCK_DGRAM, IPPROTO_TCP)) == -1)
    {
        Logger::error("ERROR opening raw socket");
    }
    
    //
    // raw socket for sending (IPPROTO_RAW == sending only)
    //
    if ((rawSockOut = socket(PF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
    {
        Logger::error("ERROR opening raw socket");
    }
    
    long bytesAvailableTunnel = 0;
    long bytesReadTunnel = 0;
    char bufferTunnel[USHRT_MAX];
    char bufferRaw[USHRT_MAX];
    unsigned short ipSizeTunnel;
    NatTranslator* nat = NatTranslator::getTranslator();
    while( bytesAvailableTunnel >= 0 && 
           !socket_is_closed(rawSockIn) )
    {
        bool action = false;
        
        /*
         *  handel packets from tunnel
         */
        bytesAvailableTunnel = bytes_on_socket(tunnelSock);
         
        // read first 4 bytes
        // byte 3 and 4 are size of ip-packet
        if( bytesReadTunnel == 0 && bytesAvailableTunnel >= 4 )
        {
            bytesReadTunnel = read(tunnelSock, bufferTunnel, 4);
            
            if( bytesReadTunnel != 4 )
                break;
            
            // grab packet size from ip header
            ipSizeTunnel = ntohs( *(unsigned short*)(bufferTunnel +2));
        }
        else if( bytesReadTunnel == 0 && socket_is_closed(tunnelSock) )
            break;
        
        // size from header has been read? 
        // read rest of packet as soon as it is fully available
        if( bytesReadTunnel == 4 && bytesAvailableTunnel +4 >= ipSizeTunnel  )
        {
            read(tunnelSock, bufferTunnel+4, ipSizeTunnel -4 );
            bytesReadTunnel = 0;
            
            if( basic_header_fields_valild(bufferTunnel, ipSizeTunnel) == false )
                break;
            struct iphdr* ip = (struct iphdr*)&bufferTunnel;
            
            nat->translateOutgoing((char*)&bufferTunnel);
            
            /*
             * send packet
             */
            Logger::log("remote put\n", Logger::VERBOSE);
            struct sockaddr_in connection;
            connection.sin_family = AF_INET;
            connection.sin_addr.s_addr = ip->daddr;  
            
            sendto(rawSockOut, 
                   bufferTunnel, 
                   ipSizeTunnel, 
                   0, 
                   (struct sockaddr *)&connection, 
                   sizeof(struct sockaddr) );
            
            action = true;
        }
        else if( bytesReadTunnel == 4 && socket_is_closed(tunnelSock) )
            break;
        
        /*
         * handle packets from raw
         */ 
        if( bytes_on_socket(rawSockIn) > 0 )
        {
            action = true;
            int bytesRead = read(rawSockIn, bufferRaw, USHRT_MAX);
            
            // ipv4 ?
           if( basic_header_fields_valild(bufferRaw, bytesRead) == false )
                continue;
           struct iphdr* ip = (struct iphdr*)&bufferRaw;
           bytesRead = ntohs(ip->tot_len); // Win7 fix
            
           // ICMP
            if(ip->protocol == IPPROTO_ICMP)
            {
                // in case of icmp only allow echo replys
                struct icmphdr* icmp = (struct icmphdr*) (bufferRaw + sizeof(struct iphdr));
                if( icmp->type != ICMP_ECHOREPLY )
                    continue;
            }
           else if(ip->protocol == IPPROTO_TCP || ip->protocol == IPPROTO_UDP)
            {
                // no further checks required
                // grab all tcp and udp traffic
            }
           else
                continue;
           
           if(ip->protocol == IPPROTO_ICMP)
           {
                // translate packet for first icmp match.
                // multiple matches are possible due to NAT.
                int total = nat->translateIncomingICMP((char*)&bufferRaw, 0);
                // at least one match found?
                if( total > 0 )
                {
     //               printf("check: 0x%x dst: 0x%x src: 0x%x reply icmp\n", ntohs(ip->check), 
     //                    ntohl(ip->daddr),
     //                    ntohl(ip->saddr));

                    Logger::log("remote grab\n", Logger::VERBOSE);

                    // send packet to tunnel
                    int bytesWritten = write(tunnelSock, bufferRaw, bytesRead);
                    // quit, if tunnel is closed
                    if( bytesWritten != bytesRead )
                        break;

                    // handle further matches, if existing
                    for(int nr=1; nr<total;nr++)
                    {
                        nat->translateIncomingICMP((char*)&bufferRaw, nr);
                        int bytesWritten = write(tunnelSock, bufferRaw, bytesRead);
                        // quit, if tunnel is closed
                        if( bytesWritten != bytesRead )
                            break;
                    }
                }
           }
           else if(ip->protocol == IPPROTO_TCP)
           {
               if( nat->translateIncomingTCP((char*)&bufferRaw) )
                {
                    //printf("check: 0x%x dst: 0x%x src: 0x%x reply tcp\n", ntohs(ip->check), 
                      //   ntohl(ip->daddr),
                        // ntohl(ip->saddr));

                    Logger::log("remote grab\n", Logger::VERBOSE);

                    int bytesWritten = write(tunnelSock, bufferRaw, bytesRead);
                    // quit, if tunnel is closed
                    if( bytesWritten != bytesRead )
                        break;
                }
           }
           else if(ip->protocol == IPPROTO_UDP)
           {
               if( nat->translateIncomingUDP((char*)&bufferRaw) )
                {
                    //printf("check: 0x%x dst: 0x%x src: 0x%x reply udp\n", ntohs(ip->check), 
                      //   ntohl(ip->daddr),
                        // ntohl(ip->saddr));

                    Logger::log("remote grab\n", Logger::VERBOSE);

                    int bytesWritten = write(tunnelSock, bufferRaw, bytesRead);
                    // quit, if tunnel is closed
                    if( bytesWritten != bytesRead )
                        break;
                }
           }
        }

        if( !action )
                usleep(1);
    }
    
    Logger::log("closed\n");
}
